kind: Role
apiVersion: rbac.authorization.k8s.io/v1
metadata:
  name: leader-election-role
  namespace: {{ name }}
rules:
- apiGroups:
  - coordination.k8s.io
  resources:
  - leases
  verbs:
  - create
  - list
  - watch
  - delete
  - get
  - patch
  - update
