[operator]
name = "{{ name }}"
# leader_acquire_interval_seconds = 10
# cleanup_interval_seconds = 5

## Webhook Server Options
# [operator.webhook_server]
# port = 443
# cert_path = "/etc/tls/tls.crt"
# key_path = "/etc/tls/tls.key"
# gunicorn_workers = 1

## Operator Metrics
# [operator.metrics]
# interval_seconds = 9
# port = 8080

# [build]
## Default builder values, you can change this to the ones that suit you
# [build.builder]
# binary = "docker"
# args = ["build", ".", "-t", "$IMG"]
## Example configuration to set image in Kustomization and build porcess
# [build.image]
# registry = "my.registry.io"
# repository = "kuroboros-operator"
# tag_prefix = "v"
# tag_suffix = ""
# tag = "$PYPROJECT_VERSION"


## Generate command related
# [generate]
# 
# ## Creates a new policy for the operator-role
# [[generate.rbac.policies]]
# api_groups = [""]
# resources = ["pods"]
# verbs = ["create", "patch", "update", "list", "watch"]