"""
Remember to run 
`kuroboros generate crd`
after modfying this file
"""
from kubernetes.client import V1ObjectMeta
from kuroboros.schema import CRDSchema, OpenAPISchema, prop


class {{ kind }}Spec(OpenAPISchema):
    """
    TODO: Define {{ kind }} spec properties
    """

class {{ kind }}Status(OpenAPISchema):
    """
    TODO: Define {{ kind }} status properties
    """

class {{ kind }}(CRDSchema):
    """ 
    TODO: Add {{ kind }} definition
    """
    kind = prop(str)
    api_version = prop(str)
    metadata = prop(V1ObjectMeta)
    
    spec = prop({{ kind }}Spec)
    status = prop({{ kind }}Status)
