import threading
from logging import Logger
from datetime import timedelta

from kuroboros.reconciler import BaseReconciler, Result
from {{ module }}.crd import {{ kind }}

class {{ kind }}Reconciler(BaseReconciler[{{ kind }}]):

    def reconcile(self, obj: {{ kind }}, stopped: threading.Event) -> Result:
        """
        TODO: Implement your reconcilation logic here
        """
        return Result(requeue=False)
