FROM python:{{ python_version }}-slim

RUN addgroup --gid 1001 app && \
    adduser \
    --disabled-password \
    --gecos "" \
    --home /home/app \
    --ingroup app \
    --uid 1001 \
    app

WORKDIR /home/app/operator
RUN chown -R app:app /home/app
ADD pyproject.toml .
RUN pip install .
ADD . .
WORKDIR /home/app/operator
USER 1001

ENTRYPOINT [ "kuroboros" ]


