from kuroboros.webhook import BaseValidationWebhook, Request, OperationsEnum

from {{ module }}.crd import {{ kind }}

class {{ kind }}Validation(BaseValidationWebhook[{{ kind }}]):

    # Modify this class variable to register your
    # webhook on this opeations
    register_on = [OperationsEnum.CREATE, OperationsEnum.UPDATE, OperationsEnum.DELETE]

    def validate(self, request: Request[{{ kind }}]) -> None:
        """
        TODO: Implement validation logic
        """
