apiVersion: v1
kind: Service
metadata:
  name: {{ name }}-metrics
  namespace: {{ name }}
  labels:
    app: {{ name }}
spec:
  type: ClusterIP
  ports:
    - port: 8080
      targetPort: 8080
      protocol: TCP
      name: metrics
  selector:
    app: {{ name }}
